/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Arrays;
import java.util.Collections;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.eingang.router.errorhandling.AdapterConfigurationException;
import de.ozgcloud.eingang.router.errorhandling.UnknownOrganisationseinheitException;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import io.grpc.Channel;
import io.grpc.stub.AbstractStub;
import net.devh.boot.grpc.client.channelfactory.GrpcChannelFactory;
import net.devh.boot.grpc.client.inject.StubTransformer;
import net.devh.boot.grpc.client.stubfactory.StubFactory;

class VorgangManagerServerResolverTest {
	@Spy
	@InjectMocks
	private VorgangManagerServerResolver resolver;

	@Mock
	private GrpcChannelFactory channelFactory;
	@Mock
	private StubFactory stubFactory;

	@Nested
	class TestCreateChannel {

		@Test
		void shouldCallChannelFactory() {
			resolver.createChannelByName(VorgangManagerListPropertiesTestFactory.VORGANG_MANAGER_NAME);

			verify(channelFactory).createChannel(VorgangManagerListPropertiesTestFactory.VORGANG_MANAGER_NAME);
		}
	}

	@Nested
	class TestGetFundstellenChannelName {

		@Nested
		class TestStrategyFundstelle {

			@Test
			void shouldCreateChannel() {
				setProperties(VorgangManagerListPropertiesTestFactory.createWithFundstelle());

				var channel = resolver.getFundstelleChannelName();

				assertThat(channel).isNotNull().isEqualTo(VorgangManagerListPropertiesTestFactory.FUNDSTELLE_CHANNEL_NAME);
			}

			@Test
			void shouldThrowExceptionIfFundstelleIsMissing() {
				var props = VorgangManagerListPropertiesTestFactory.createWithFundstelle();
				props.setFundstelleVorgangManagerName(Optional.empty());
				setProperties(props);

				assertThrows(AdapterConfigurationException.class, () -> resolver.getFundstelleChannelName());
			}
		}

		@Nested
		class TestStrategyDeny {

			@Test
			void shouldThrowException() {
				setProperties(VorgangManagerListPropertiesTestFactory.createForMultiRouting());

				assertThrows(UnknownOrganisationseinheitException.class, () -> resolver.getFundstelleChannelName());
			}
		}
	}

	@Nested
	class TestGetChannelName {

		@Test
		void shouldUseSingleName() {
			setProperties(VorgangManagerListPropertiesTestFactory.createForSingleRouting());

			var name = resolver.getChannelName(Optional.empty());

			assertThat(name).contains(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME);
		}

		@Test
		void shouldUseNameFromMap() {
			setProperties(VorgangManagerListPropertiesTestFactory.createForMultiRouting());

			var name = resolver.getChannelName(Optional.of(VorgangManagerListPropertiesTestFactory.ORGANISATIONSEINHEIT_ID));

			assertThat(name).contains(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME);
		}

		@Test
		void shouldGetFundstellenName() {
			setProperties(VorgangManagerListPropertiesTestFactory.createWithFundstelle());

			var name = resolver.getChannelName(Optional.of("4711"));

			verify(resolver).getFundstelleChannelName();
			assertThat(name).isEqualTo(VorgangManagerListPropertiesTestFactory.FUNDSTELLE_CHANNEL_NAME);
		}
	}

	@Nested
	class TestFindStubFactory {

		@Mock
		private StubFactory stubFactory;

		@Test
		void shouldSetApplicableFactory() {
			when(stubFactory.isApplicable(any())).thenReturn(true);
			setStubFactories(stubFactory, stubFactory);

			resolver.findApplicableStubFactories();

			assertThat(ReflectionTestUtils.getField(resolver, "vorgangStubFactory")).isSameAs(stubFactory);
		}

		@Test
		void shouldThrowExceptionIfNotFound() {
			setStubFactories(stubFactory);

			assertThrows(AdapterConfigurationException.class, () -> resolver.findApplicableStubFactories());
		}

		private void setStubFactories(StubFactory... factories) {
			ReflectionTestUtils.setField(resolver, "stubFactories", Arrays.asList(factories));
		}
	}

	@Nested
	class TestCreateStub {

		@Mock
		private Channel channel;
		@Mock
		private StubFactory stubFactory;

		private Class<? extends AbstractStub<?>> stubClass = VorgangServiceBlockingStub.class;

		@BeforeEach
		void initTest() {
			doReturn(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME).when(resolver).getChannelName(any());
			doReturn(channel).when(resolver).createChannelByName(any());
		}

		@Test
		void shouldGetChannel() {
			createStub();

			verify(resolver).createChannelByName(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME);
		}

		@Test
		void shouldApplyTransformers() {
			createStub();

			verify(resolver).applyStubTransformers(any(), any());
		}

		@Test
		void shouldCreateStubByFactory() {
			createStub();

			verify(stubFactory).createStub(eq(stubClass), any());
		}

		private AbstractStub<?> createStub() {
			return resolver.createStub(Optional.of(VorgangManagerListPropertiesTestFactory.ORGANISATIONSEINHEIT_ID), stubFactory, stubClass);
		}
	}

	@Nested
	class TestApplyStubTransformer {

		@Mock
		private AbstractStub<?> stub;
		@Mock
		private StubTransformer transformer;

		@BeforeEach
		void init() {
			ReflectionTestUtils.setField(resolver, "stubTransformers", Collections.singleton(transformer));
		}

		@Test
		void shouldCallTransform() {
			resolver.applyStubTransformers(stub, VorgangManagerListPropertiesTestFactory.CHANNEL_NAME);

			verify(transformer).transform(VorgangManagerListPropertiesTestFactory.CHANNEL_NAME, stub);
		}
	}

	private void setProperties(VorgangManagerListProperties properties) {
		ReflectionTestUtils.setField(resolver, "properties", properties);
	}
}
