package de.ozgcloud.eingang.xta;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;

import javax.net.ssl.KeyManagerFactory;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.webservices.client.WebServiceTemplateCustomizer;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.oxm.jaxb.Jaxb2Marshaller;
import org.springframework.ws.client.support.destination.DestinationProvider;
import org.springframework.ws.client.support.interceptor.ClientInterceptor;
import org.springframework.ws.soap.SoapVersion;
import org.springframework.ws.soap.saaj.SaajSoapMessageFactory;
import org.springframework.ws.transport.WebServiceMessageSender;
import org.springframework.ws.transport.http.HttpsUrlConnectionMessageSender;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Configuration
public class XtaRemoteServiceConfiguration {

	static final String URI_TEMPLATE = "%s://%s/MB_XTA-WS/XTA210msgBoxPort.svc";

	@Autowired
	private XtaProperties properties;

	@Bean
	Jaxb2Marshaller osciTransportMarshaller() {
		Jaxb2Marshaller marshaller = new Jaxb2Marshaller();
		marshaller.setContextPath("eu.osci.ws._2008._05.transport");
		marshaller.setMtomEnabled(true);
		return marshaller;
	}

	@Bean
	Jaxb2Marshaller xoevTransportMarshaller() {
		Jaxb2Marshaller unmarshaller = new Jaxb2Marshaller();
		unmarshaller.setContextPath("de.xoev.transport.xta._211");
		unmarshaller.setMtomEnabled(true);
		return unmarshaller;
	}

	@Bean
	WebServiceTemplateCustomizer webServiceTemplateCustomizer() {
		return template -> template.setMessageSender(messageSender());
	}

	@Bean
	WebServiceTemplateCustomizer setMessageFactoryCustomizer() {
		return template -> template.setMessageFactory(messageFactory());
	}

	@Bean
	WebServiceTemplateCustomizer addingInterceptorCustomizer(WsHeaderAddingInterceptor interceptor) {
		return template -> template.setInterceptors(new ClientInterceptor[] { interceptor });
	}

	@Bean
	WebServiceTemplateCustomizer setDestionationProvider() {
		return template -> template.setDestinationProvider(destinationProvider());
	}

	@Bean
	DestinationProvider destinationProvider() {
		return () -> {
			try {
				String serverUri = buildServerAddressUri(properties.getServer());
				LOG.trace("Xta Service remote URI: {}", serverUri);
				return new URI(serverUri);
			} catch (URISyntaxException e) {
				throw new TechnicalException("Error building URI", e);
			}
		};
	}

	String buildServerAddressUri(Server server) {
		return URI_TEMPLATE.formatted(server.getProtocol(), server.getAddress());
	}

	@Bean
	SaajSoapMessageFactory messageFactory() {
		SaajSoapMessageFactory messageFactory = new SaajSoapMessageFactory();
		messageFactory.setSoapVersion(SoapVersion.SOAP_12);
		return messageFactory;
	}

	@Bean
	WebServiceMessageSender messageSender() {
		try {
			var keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
			keyManagerFactory.init(xtaKeyStore(), properties.getKeyStore().getPassword());

			var messageSender1 = new HttpsUrlConnectionMessageSender();
			messageSender1.setKeyManagers(keyManagerFactory.getKeyManagers());
			messageSender1.setHostnameVerifier((hostname, session) -> true); // NOSONAR hostname verification is senseless due missing DNS for
																				// Dataport XTA Server

			return messageSender1;
		} catch (Exception e) {
			throw new TechnicalException("Error initializating message sender.", e);
		}
	}

	@Bean
	KeyStore xtaKeyStore() throws KeyStoreException, CertificateException, NoSuchAlgorithmException, IOException {
		var keyStoreResource = properties.getKeyStore().getFile();
		var keyStore = KeyStore.getInstance(properties.getKeyStore().getType());
		try (InputStream keyStoreStream = keyStoreResource.getInputStream()) {
			keyStore.load(keyStoreStream, properties.getKeyStore().getPassword());
		}

		return keyStore;
	}

}
